use super::hypixel::HypixelResponse;
use crate::Caches;
use crate::Error::{self, *};
use poise::serenity_prelude::User;
use reqwest::Client;
use sqlx::query_as;
use sqlx::Pool;
use sqlx::Sqlite;

#[derive(PartialEq, sqlx::FromRow)]
pub(crate) struct Uuid {
    uuid: String,
}

impl Uuid {
    pub(crate) fn get(self) -> String {
        self.uuid
    }

    pub(crate) fn as_str(&self) -> &str {
        self.uuid.as_str()
    }

    pub(crate) async fn for_ign(ign: &str, cli: &Client, c: &Caches) -> Result<Self, Error> {
        let uuid = crate::data::mojang::uuid(c, cli, ign.to_owned()).await?;
        Ok(Self { uuid })
    }

    pub(crate) async fn ign(&self, c: &Caches, cli: &Client) -> Result<String, Error> {
        let ign = crate::data::mojang::name(c, cli, self.uuid.clone()).await?;
        Ok(ign)
    }
}

impl Uuid {
    pub(crate) async fn has_discord_user(
        &self,
        user: &User,
        client: &Client,
    ) -> Result<bool, Error> {
        let res = HypixelResponse::get(self.uuid.as_str(), client).await?;
        let matches = res.discord().ok_or(Other(
            "The Hypixel profile has no Discord account linked. Please follow the steps in \
             <#1256219552568840263>"
                .to_string(),
        ))? == user.name;
        Ok(matches)
    }
}

#[derive(PartialEq)]
pub(crate) struct DiscordId {
    inner: u64,
}

impl<'a, R: sqlx::Row> sqlx::FromRow<'a, R> for DiscordId
where
    &'a ::std::primitive::str: sqlx::ColumnIndex<R>,
    i64: ::sqlx::decode::Decode<'a, R::Database>,
    i64: ::sqlx::types::Type<R::Database>,
{
    fn from_row(row: &'a R) -> sqlx::Result<Self> {
        let discord_id: i64 = row.try_get("discord_id")?;
        Ok(DiscordId {
            inner: discord_id.cast_unsigned(),
        })
    }
}

impl std::fmt::Display for DiscordId {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.inner)
    }
}

pub(crate) struct Link {
    link_id: LinkId,
    discord_ids: Vec<DiscordId>,
    minecraft_accounts: Vec<Uuid>,
}

impl From<Link> for Vec<Uuid> {
    fn from(value: Link) -> Self {
        value.minecraft_accounts
    }
}

impl Link {
    pub(crate) fn minecraft_accounts(&self) -> &Vec<Uuid> {
        &self.minecraft_accounts
    }

    pub(crate) fn discord_ids(&self) -> &Vec<DiscordId> {
        &self.discord_ids
    }

    pub(crate) fn link_id(&self) -> LinkId {
        self.link_id
    }

    async fn lookup_by_id(link_id: LinkId, pool: &Pool<Sqlite>) -> Result<Link, Error> {
        let link = Link {
            link_id,
            discord_ids: query_as(
                format!("SELECT discord_id FROM discord_links WHERE link_id = {link_id};").as_str(),
            )
            .fetch_all(pool)
            .await?,
            minecraft_accounts: query_as(
                format!(
                    "SELECT minecraft_uuid AS uuid FROM minecraft_links WHERE link_id = {link_id};"
                )
                .as_str(),
            )
            .fetch_all(pool)
            .await?,
        };
        Ok(link)
    }

    pub(crate) async fn try_from_discord(user: &User, pool: &Pool<Sqlite>) -> Result<Self, Error> {
        let link_id = LinkId::try_from_discord(pool, user.id.get()).await?;
        Link::lookup_by_id(link_id, pool).await
    }

    /* #[allow(dead_code)]
     * pub(crate) async fn try_from_minecraft(
     *   player: Uuid,
     *   pool: &Pool<Sqlite>,
     * ) -> Result<Self, Error> {
     *   let link_id = LinkId::try_from_minecraft(pool, player.uuid.as_str()).await?;
     *   Link::lookup_by_id(link_id, pool).await
     * }
    */
}

#[derive(Clone, Copy)]
pub(crate) struct LinkId {
    inner: u16,
}

impl<'a, R: sqlx::Row> sqlx::FromRow<'a, R> for LinkId
where
    &'a ::std::primitive::str: sqlx::ColumnIndex<R>,
    i16: ::sqlx::decode::Decode<'a, R::Database>,
    i16: ::sqlx::types::Type<R::Database>,
{
    fn from_row(row: &'a R) -> sqlx::Result<Self> {
        let link_id: i16 = row.try_get("link_id")?;
        Ok(Self {
            inner: link_id.cast_unsigned(),
        })
    }
}

impl std::fmt::Display for LinkId {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.inner)
    }
}

impl LinkId {
    fn incremented(self) -> Self {
        Self {
            inner: self.inner + 1,
        }
    }

    pub(crate) async fn new(pool: &Pool<Sqlite>) -> Result<LinkId, Error> {
        let highest: LinkId = query_as("SELECT MAX(link_id) AS link_id FROM minecraft_links;")
            .fetch_one(pool)
            .await?;
        Ok(highest.incremented())
    }

    pub(crate) async fn try_from_discord(
        pool: &Pool<Sqlite>,
        snowflake: u64,
    ) -> Result<LinkId, Error> {
        query_as(
            format!(
                "SELECT link_id FROM discord_links WHERE discord_id = {} LIMIT 1;",
                snowflake.cast_signed()
            )
            .as_str(),
        )
        .fetch_optional(pool)
        .await?
        .ok_or(Other("This user has no accounts linked.".to_string()))
    }

    pub(crate) async fn try_from_minecraft(
        pool: &Pool<Sqlite>,
        minecraft_uuid: &str,
    ) -> Result<LinkId, Error> {
        query_as(
        format!(
            r#"SELECT link_id FROM minecraft_links WHERE minecraft_uuid = "{minecraft_uuid}" LIMIT 1;"#
        )
        .as_str(),
    )
    .fetch_optional(pool)
    .await?
    .ok_or(Other("This player has no accounts linked.".to_string()))
    }
}
